"""
Parse ASCII PSF Files
"""

# Imports {{{1
import ply.lex
import ply.yacc
from inform import Info, is_str, is_mapping


# Globals {{{1
Filename = None


# Utility classes {{{1
class Type(Info):
    pass


class Struct(Info):
    pass


class Sweep(Info):
    pass


class Trace(Info):
    pass


class Traces(Info):
    pass


class Value(Info):
    pass


# Exceptions {{{1
# ParseError {{{2
class ParseError(Exception):
    def __init__(self, text, loc=None):
        self.text = text
        self.loc = loc

    def __str__(self):
        "Return a formatted error message."
        if self.loc:
            return self.loc.message(Filename, self.text)
        if Filename:
            return "%s: %s" % (Filename, self.text)
        return self.text


# Record the location of a token {{{2
class TokenLocation(object):
    def __init__(self, token, index=None):
        "Records information about the location of a token."
        lexdata = token.lexer.lexdata
        if index:
            lexpos = token.lexpos(index)
        else:
            lexpos = token.lexpos
        bol = lexdata.rfind('\n', 0, lexpos) + 1
        if bol < 0:
            bol = 0
        eol = lexdata.find('\n', lexpos)
        if eol < 0:
            eol = len(lexdata)
        self.line = lexdata[bol:eol]
        self.col = (lexpos - bol)
        self.row = lexdata[0:lexpos].count('\n') + 1

    def annotateLine(self, prefix=None):
        """
        Produces a two or three line result, possibly a prefix, then the line
        that contains the token, and then a pointer to the token.  If a prefix
        is given, it will be printed before the line and will be separated from
        the line by ': '.  Generally the prefix contains the filename and line
        number.
        """
        if prefix:
            return "%s\n    %s\n    %s^" % (
                prefix,
                self.line,
                self.col*' '
            )
        return "%s\n%s^" % (self.line, (self.col-1)*' ')

    def message(self, filename, msg):
        """
        Produces a message about the token.  The line containing the token is
        shown, along with a pointer to the token.  Finally the message is
        returned.
        """
        # if self.row:
        #     loc = "%s.%s" % (self.row, self.col)
        if self.row:
            loc = "%s" % (self.row)
        else:
            loc = ""
        if filename:
            loc = "%s(%s)" % (filename, loc)
        if loc:
            return self.annotateLine("%s: %s" % (loc, msg))
        return self.annotateLine(msg)


# Lexer {{{1
# List of the token names
reserved = {rw: rw for rw in [
    'BYTE',
    'COMPLEX',
    'DOUBLE',
    'END',
    'FLOAT',
    'GROUP',
    'HEADER',
    'INT',
    'LONG',
    'NAN',
    'PROP',
    'STRUCT',
    'SWEEP',
    'TRACE',
    'TYPE',
    'VALUE',
]}
tokens = [
    'INTEGER',
    'REAL',
    'STRING',
] + list(reserved.values())

# Literal tokens
literals = r'()'

# Regular expression tokens
# Regular expressions that define numbers
t_INTEGER = r"-?[0-9]+"
t_REAL = r"[+-]?[0-9]+\.[0-9]*([eE][+-][0-9]+)?"
t_NAN = r"nan|NaN"

# Regular expression for a string
t_STRING = r'"[^\n"]*"'


# Identifiers
def t_ID(t):
    r'[A-Z]+'
    t.type = reserved.get(t.value)
    if t.type is None:
        loc = TokenLocation(t)
        t.lexer.skip(1)
        raise ParseError(f"unknown keyword '{t.value}'.", loc)
    return t


# ignore whitespace
t_ignore = ' \t\n'


def t_error(t):
    c = t.value[0]
    loc = TokenLocation(t)
    t.lexer.skip(1)
    raise ParseError("illegal character '%s'." % c, loc)


# Parser {{{1
def p_contents(p):
    "contents : header_section type_section sweep_section trace_section value_section end"
    p[0] = (p[1], p[2], p[3], p[4], p[5])


def p_contents_without_sweep(p):
    "contents : header_section type_section value_section end"
    p[0] = (p[1], p[2], None, None, p[3])


def p_header_section(p):
    "header_section : HEADER named_values"
    p[0] = dict(p[2])


def p_named_values(p):
    "named_values : named_values named_value"
    p[1].append(p[2])
    p[0] = p[1]


def p_named_values_last(p):
    "named_values : named_value"
    p[0] = [p[1]]


def p_named_value(p):
    "named_value : string value"
    p[0] = (p[1], p[2])


def p_string_value(p):
    "value : string"
    p[0] = p[1]


def p_integer_value(p):
    "value : INTEGER"
    p[0] = int(p[1])


def p_real_value(p):
    """
    value : REAL
          | NAN
    """
    p[0] = float(p[1])


def p_string(p):
    "string : STRING"
    p[0] = (p[1][1:-1]).replace('\\', '')


def p_type_section(p):
    "type_section : TYPE types"
    p[0] = dict(p[2])


def p_types(p):
    "types : types type"
    p[1].append(p[2])
    p[0] = p[1]


def p_types_last(p):
    "types : type"
    p[0] = [p[1]]


def p_type(p):
    "type : string kinds"
    n = p[1]
    meta = {}
    kind = ' '.join(s for s in p[2] if is_str(s)).lower()
    if kind:
        meta['kind'] = kind
    for each in p[2]:
        if is_mapping(each):
            meta.update(each)
        elif isinstance(each, Struct):
            meta['struct'] = each
    p[0] = (n, Type(name=n, **meta))


def p_kinds(p):
    "kinds : kinds kind"
    p[1].append(p[2])
    p[0] = p[1]


def p_kinds_last(p):
    "kinds : kind"
    p[0] = [p[1]]


def p_kind(p):
    """
        kind : FLOAT
             | DOUBLE
             | COMPLEX
             | INT
             | BYTE
             | LONG
             | struct
             | prop
    """
    p[0] = p[1]


def p_struct(p):
    "struct : STRUCT '(' types ')'"
    p[0] = Struct(types=dict(p[3]))


def p_prop(p):
    "prop : PROP '(' named_values ')'"
    p[0] = dict(p[3])


def p_sweep_section(p):
    "sweep_section : SWEEP sweeps"
    p[0] = p[2]


def p_sweeps(p):
    "sweeps : sweeps sweep"
    p[1].append(p[2])
    p[0] = p[1]


def p_sweeps_last(p):
    "sweeps : sweep"
    p[0] = [p[1]]


def p_sweep(p):
    "sweep : string string kinds"
    p[0] = Sweep(name=p[1], type=p[2], **p[3][0])


def p_trace_section(p):
    "trace_section : TRACE traces"
    # Partition out the groups from the traces.
    # A group will have one entry in the list of traces, and it will have a
    # corresponding entry in the groups directory.  The entry is also a
    # dictionary that maps the member name to the member type.
    traces = []
    groups = {}
    index = None
    for trace in p[2]:
        name = trace.name
        try:
            count = int(trace.type)
            index = 0
            group_name = name
            groups[group_name] = {}
            trace.type = 'GROUP'
            traces.append(trace)
            continue
        except ValueError:
            pass
        if index is None:
            traces.append(trace)
        else:
            groups[group_name][trace.name] = trace.type
            index += 1
            if index == count:
                index = None
    p[0] = (traces, groups)


def p_traces(p):
    "traces : traces trace"
    p[1].append(p[2])
    p[0] = p[1]


def p_traces_last(p):
    "traces : trace"
    p[0] = [p[1]]


def p_trace(p):
    "trace : named_value"
    name, type = p[1]
    p[0] = Trace(name=name, type=type)


def p_group_trace(p):
    "trace : string GROUP INTEGER"
    name = p[1]
    count = p[3]
    p[0] = Trace(name=name, type=count)


def p_trace_with_props(p):
    "trace : named_value prop"
    # some psf files place a units property on terminal current traces,
    # but the information seems redundant and can be ignored.
    name, type = p[1]
    p[0] = Trace(name=name, type=type)


def p_value_section(p):
    "value_section : VALUE values"
    values = {}
    for n, t, v in p[2]:
        if n not in values:
            values[n] = Value(type=t, values=[])
        values[n].values.append(v)
    p[0] = values


def p_values(p):
    "values : values signal_value"
    p[1].append(p[2])
    p[0] = p[1]


def p_values_last(p):
    "values : signal_value"
    p[0] = [p[1]]


def p_named_signal_scalar(p):
    """
    signal_value : string numbers
    """
    p[0] = (p[1], None, p[2])


def p_named_signal_with_type(p):
    """
    signal_value : string string numbers
    """
    p[0] = (p[1], p[2], p[3])


def p_integer_number(p):
    "simple_number : INTEGER"
    p[0] = int(p[1])


def p_real_number(p):
    # props are redundant, so ignore them
    """
    simple_number : REAL
                  | REAL prop
                  | NAN
                  | NAN prop
    """
    p[0] = float(p[1])


def p_simple_numbers(p):
    "simple_numbers : simple_numbers simple_number"
    p[1].append(p[2])
    p[0] = p[1]


def p_simple_numbers_last(p):
    "simple_numbers : simple_number"
    p[0] = [p[1]]


def p_composite_number(p):
    """
    composite_number : '(' simple_numbers ')'
                     | '(' simple_numbers ')' prop
    """
    p[0] = tuple(p[2])


def p_composite_numbers(p):
    "composite_numbers : composite_numbers composite_number"
    p[1].append(p[2])
    p[0] = p[1]


def p_composite_numbers_last(p):
    "composite_numbers : composite_number"
    p[0] = [p[1]]


def p_numbers(p):
    """
    numbers : simple_numbers
            | composite_numbers
    """
    p[0] = p[1]


def p_end(p):
    "end : END"


# Error rule for syntax errors
def p_error(p):
    if p:
        loc = TokenLocation(p)
        raise ParseError("syntax error at '%s'." % (p.value), loc)
    else:
        raise ParseError("premature end of content.")


# ParsePSF class {{{1
class ParsePSF:
    def __init__(self):
        self.lexer = ply.lex.lex()
        self.parser = ply.yacc.yacc(write_tables=False, debug=False)

    def parse(self, filename, content):
        global Filename
        Filename = filename

        result = self.parser.parse(content, tracking=False, lexer=self.lexer)
        return result



# Usage {{{1
"""
Show Signals

Signals may either be waveforms or single points.
Waveforms are plotted and single points are printed.

Usage:
    show-psf [options] <signal>...

Options:
    -c, --no-cache                ignore, then regenerate, the cache
    -f <path>, --psf-file <path>  PSF file
    -d, --db                      show the magnitude of the signals in dB
    -m, --mag                     show the magnitude of the signals
    -p, --ph                      show the phase of the signals
    -s <file>, --svg <file>       produce graph as SVG file rather than display it
    -t <title>, --title <title>   title
    -M, --mark-points             place marker on each point
    -P, --just-points             do not connect points with lines (implies -M)

The PSF file need only be given if it differs from the one used previously.

Reading large ASCII data files is slow, so show-psf reads the PSF file once,
then pickles the data and writes it to disk. On subsequent runs the pickled data
is used if the pickle file is newer that the corresponding PSF file.

A signal may contain glob characters. For examples, R1:* shows all signals that
start with R1:.

If a signal as specified on the command line contains a dash, it is split into
two pieces, each of which are considered signals that are components of a
differential signal. The two are accessed individually and the difference is
shown. So for example, out_p-out_n results in V(out_p, ount_n) being shown.
There may only be one dash in a signal, and signals with dashes must not contain
glob characters.
"""


# Imports {{{1
from .psf import PSF, Quantity
from docopt import docopt
import fnmatch
from inform import (
    Error, display, done, fatal, os_error, plural, warn
)
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
import numpy as np
import warnings
import sys

# Globals {{{1
axis_prec = 9  # can request many digits as unneeded digits are not shown
cursor_prec = 3
print_prec = 4
Quantity.set_prefs(
    map_sf = Quantity.map_sf_to_sci_notation,
    minus = Quantity.minus_sign,
    # include small scale factors for noise power results
    output_sf = 'TGMkmunpfazy',
    prec = axis_prec,
)
warnings.filterwarnings('ignore', category=FutureWarning)
saved_psf_file_filename = '.psf_file'
saved_arguments_filename = '.psf_show_args'
operators = '+ - * /'.split()


# Utilities {{{1
# get_argv() {{{3
def get_argv():
    argv = sys.argv[1:]
    if argv:
        # save the command line arguments for next time
        try:
            with open(saved_arguments_filename, 'w') as f:
                args = [a for a in argv if a not in ['-c', '--no-cache']]
                f.write('\n'.join(args))
        except OSError as e:
            warn(os_error(e))
    else:
        # command line arguments not give, reuse previous ones
        try:
            with open(saved_arguments_filename) as f:
                argv = f.read().split('\n')
            display('Using command:', ' '.join(argv))
        except OSError:
            done()
    return argv


# get_psf_filename() {{{2
def get_psf_filename(psf_file):
    if not psf_file:
        try:
            with open(saved_psf_file_filename) as f:
                psf_file = f.read().strip()
            display('Using PSF file:', psf_file)
        except OSError:
            fatal('missing PSF file name.')
    try:
        with open(saved_psf_file_filename, 'w') as f:
            f.write(psf_file)
    except OSError as e:
        warn(os_error(e))
    return psf_file


# in_args() {{{2
def expand_args(signals, args, allow_diff=True):
    # special case args that contain -, they are considered differential signals
    # they should not include glob chars (*, ?)
    selected = set(a for a in args if '-' in a) if allow_diff else set()
    for arg in args:
        selected.update(fnmatch.filter(signals, arg))
    return sorted(selected)


# show_signals() {{{1
def show_signals():
    try:
        # process command line {{{2
        cmdline = docopt(__doc__, argv=get_argv())
        psf_file = get_psf_filename(cmdline['--psf-file'])
        args = cmdline['<signal>']
        title = cmdline['--title']
        svg_file = cmdline['--svg']
        dB = cmdline['--db']
        mag = cmdline['--mag']
        phase = cmdline['--ph']
        use_cache = not cmdline['--no-cache']
        linestyle = '' if cmdline['--just-points'] else '-'
        marker = '.' if cmdline['--mark-points'] or cmdline['--just-points'] else ''

        # Open PSF file {{{2
        psf = PSF(psf_file, sep=':', use_cache=use_cache)
        sweep = psf.get_sweep()
        to_show = expand_args(psf.signals.keys(), args)

        # Print scalars {{{2
        if not sweep:
            with Quantity.prefs(map_sf = Quantity.map_sf_to_greek, prec = print_prec):
                to_print = []
                width = 0
                for arg in to_show:
                    pair = arg.split('-')
                    if len(pair) == 2:
                        psig = psf.get_signal(pair[0])
                        nsig = psf.get_signal(pair[1])
                        if psig.units != nsig.units:
                            warn(
                                f'incompatible units ({psig.units} != {nsig.units}',
                                culprit=arg
                            )
                        units = psig.units
                        access = psig.access
                        y_data = Quantity(psig.ordinate - nsig.ordinate, units)
                        if access:
                            name = f'{access}({psig.name},{nsig.name})'
                        else:
                            name = f'({psig.name} − {nsig.name})'
                    else:
                        sig = psf.get_signal(arg)
                        access = sig.access
                        name = f'{access}({sig.name})' if access else sig.name
                        y_data = sig.ordinate
                    to_print.append((name, y_data))
                    width = max(width, len(name))
                for name, y_data in to_print:
                    display(f'{name:>{width+4}} = {y_data}')
                return

        # Process arguments for graphs {{{2
        waves = []
        y_units = set()
        for arg in to_show:
            use_log_scale = psf.log_y(sweep)
            pair = arg.split('-')
            if len(pair) == 2:
                psig = psf.get_signal(pair[0])
                nsig = psf.get_signal(pair[1])
                name = arg
                if psig.units != nsig.units:
                    warn(
                        f'incompatible units ({psig.units} != {nsig.units}',
                        culprit=arg
                    )
                units = psig.units
                y_data = psig.ordinate - nsig.ordinate
            else:
                sig = psf.get_signal(arg)
                name = arg
                units = sig.units
                y_data = sig.ordinate
            if units == 'Unitless':
                units = ''
            if dB:
                y_data = 20*np.log10(np.absolute(y_data))
                units = 'dB' + units
                use_log_scale = False
            elif mag:
                y_data = np.absolute(y_data)
            elif phase:
                y_data = np.angle(y_data, deg=True)
                units = '°'
                use_log_scale = False
            elif np.iscomplexobj(y_data):
                y_data = np.absolute(y_data)
            waves.append((name, y_data, units, use_log_scale))
            y_units.add(units)
        if not y_units:
            raise Error(f'{plural(args):no match/es}.', culprit=args)

        # Formatters {{{2
        # create formatter for x-axis values {{{3
        x_units = sweep.units
        x_data = sweep.abscissa
        x_formatter = FuncFormatter(
            lambda v, p: Quantity(v, x_units).render()
        )

        # create formatter for y-axis values {{{3
        y_formatters = {
            u: FuncFormatter(
                lambda v, p, u=u: str(Quantity(v, psf.units_to_unicode(u)))
            )
            for u in y_units
        }

        # create formatter for cursor readout values {{{3
        xy_formatters = {}
        for u in y_units:
            xy_label = "{x},  {y}"
            units = psf.units_to_unicode(u)
            xy_formatters[u] = lambda x, y: xy_label.format(
                x = Quantity(x, x_units).render(prec=cursor_prec),
                y = Quantity(y, units).render(prec=cursor_prec)
            )

        # Generate the graph {{{2
        if svg_file:
            matplotlib.use('SVG')
        figure, axes = plt.subplots(len(y_units), 1, sharex=True, squeeze=False)
        for i, units in enumerate(y_units):
            for sig_name, y_data, sig_units, use_log_scale in waves:
                if sig_units == units:
                    axes[i, 0].plot(
                        x_data, y_data,
                        label = sig_name,
                        marker = marker,
                        linestyle = linestyle,
                        linewidth = 2,
                    )
            axes[i, 0].legend(frameon=False, loc='best')
            axes[i, 0].set_xscale('log' if psf.log_x(sweep) else 'linear')
            axes[i, 0].set_yscale('log' if use_log_scale else 'linear')
            axes[i, 0].xaxis.set_major_formatter(x_formatter)
            axes[i, 0].yaxis.set_major_formatter(y_formatters[units])
            axes[i, 0].format_coord = xy_formatters[units]
        if title:
            plt.suptitle(title)
        if svg_file:
            plt.savefig(svg_file)
        else:
            plt.show()
    except Error as e:
        e.terminate()
    except OSError as e:
        fatal(os_error(e))
    except KeyboardInterrupt:
        done()



#!/usr/bin/env python3
# Usage {{{1
"""
List Signals

Usage:
    list-psf [options] [<signal>...]

Options:
    -c, --no-cache                ignore, then regenerate, the cache
    -f <path>, --psf-file <path>  the path of the ASCII PSF file
    -l, --long                    include signal meta data

The PSF file need only be given if it differs from the one use previously.

Reading large ASCII data files is slow, so list-psf reads the PSF file once,
then pickles the data and writes it to disk. On subsequent runs the pickled data
is used if the pickle file is newer that the corresponding PSF file.
"""

# Imports {{{1
from .show import expand_args, get_psf_filename
from .psf import PSF
from docopt import docopt
from inform import Error, columns, display, plural, warn
import warnings

# Globals {{{1
warnings.filterwarnings('ignore', category=FutureWarning)
saved_psf_file_filename = '.psf_file'
kinds = {
    'float double': 'real',
    'float complex': 'complex',
    'int byte': 'integer',
    'int long': 'integer',
    'int': 'integer',
}


# list_signals() {{{1
def list_signals():
    # Read command line {{{2
    cmdline = docopt(__doc__)
    args = cmdline['<signal>']
    if not args:
        args = ['*']
    psf_file = get_psf_filename(cmdline['--psf-file'])
    show_meta = cmdline['--long']
    use_cache = not cmdline['--no-cache']

    # List signals {{{2
    try:
        psf = PSF(psf_file, sep=':', use_cache=use_cache)

        if show_meta:
            nw = uw = kw = 0  # name width, units width, kind width
            data = []
            for name in expand_args(psf.signals.keys(), args, allow_diff=False):
                if name not in psf.signals:
                    warn('not found.', culprit=name)
                signal = psf.get_signal(name)
                if len(signal.name) > nw:
                    nw = len(signal.name)
                units = psf.units_to_unicode(signal.units)
                if len(units) > uw:
                    uw = len(units)
                kind = signal.type.kind
                kind = kinds.get(kind, kind)
                if len(kind) > kw:
                    kw = len(kind)
                try:
                    points = len(signal.ordinate)
                except TypeError:
                    points = None
                data.append((signal.name, units, kind, points))
            if not data:
                raise Error(f'{plural(args):no match/es}.', culprit=args)
            for name, units, kind, points in data:
                if points is None:
                    display(f'    {name:<{nw}}  {units:<{uw}}  {kind}')
                else:
                    display(f'    {name:<{nw}}  {units:<{uw}}  {kind:<{kw}}  ({points} points)')
        else:
            signals = expand_args(psf.signals.keys(), args, allow_diff=False)
            if not signals:
                raise Error(f'{plural(args):no match/es}.', culprit=args)
            display(columns(signals))
    except Error as e:
        e.terminate()



"""
Read PSF File
"""

# Imports {{{1
from .parse import ParsePSF, ParseError
from inform import Error, Info, join, log, os_error
from pathlib import Path
import numpy as np
from quantiphy import Quantity
try:
    import cPickle as pickle
except ImportError:
    import pickle
import re


# Utilities {{{1
class Signal(Info):
    pass


class UnknownSignal(Error):
    template = 'unknown signal: {}.'


unicode_unit_maps = {
    r'sqrt\(([^)]+)\)': r'√\1',
    r'\^2': '²',
    r'\bOhm\b': 'Ω',

    # the following are not unicode units, however Spectre often messes these
    # up in the oppoint files.
    r'\bR\b': 'Ω',
    r'\bI\b': 'A',
    r'\bC\b': 'F',
}


def unicode_units(u):
    if u:
        for s, r in unicode_unit_maps.items():
            u = re.sub(s, r, u)
    else:
        u = ''
    return u


# PSF class {{{1
class PSF:
    """
    Read an ASCII PSF file

    filename (str or Path):
        Path to ASCII PSF file.
    sep (str):
        Join string to use when converting composite names into a single name.
    use_cache (bool):
        If True, a cached version of the data is used if it exists rather than
        the original PSF file, but only if the cache is newer than the PSF file.
        This can substantially reduced the time required to access the data.
    update_cache (bool):
        If True, a cached version of the data is updated if it does not exist or
        is out-of-date.
    """

    def __init__(self, filename, sep=':', use_cache=True, update_cache=True):
        psf_filepath = Path(filename)
        cache_filepath = psf_filepath.with_suffix(psf_filepath.suffix + '.cache')

        # read cache if desired and current
        if use_cache:
            try:
                if cache_filepath.stat().st_mtime > psf_filepath.stat().st_mtime:
                    self._read_cache(cache_filepath)
                    return
            except OSError as e:
                log(os_error(e))
            except Exception as e:
                log(e)

        # open and parse PSF file
        parser = ParsePSF()
        try:
            content = psf_filepath.read_text()
            sections = parser.parse(filename, content)
        except ParseError as e:
            raise Error(str(e))
        except OSError as e:
            raise Error(os_error(e))
        except UnicodeError as e:
            raise Error(
                e,
                culprit = psf_filepath,
                codicil = join(
                    'This is likely a binary PSF file,',
                    'psf_utils only supports ASCII PSF files.',
                    '\nUse `psf {0!s} {0!s}.ascii` to convert.'.format(psf_filepath),
                )
            )
        meta, types, sweeps, traces, values = sections
        self.meta = meta
        self.types = types
        self.sweeps = sweeps
        self.traces = traces

        # add values to sweeps
        if sweeps:
            for sweep in sweeps:
                n = sweep.name
                sweep.abscissa = np.array([v[0] for v in values[n].values])

        # process signals
        # 1. convert to numpy and delete the original list
        # 2. convert to Signal class
        # 3. create signals dictionary
        signals = {}
        if traces:
            traces, groups = traces
            for trace in traces:
                name = trace.name
                type = types.get(trace.type, trace.type)
                vals = values[name].values
                if type == 'GROUP':
                    group = {k: types.get(v, v) for k, v in groups[name].items()}
                    prefix = ''
                    get_value = lambda v, i: v[i]
                elif type.struct:
                    group = type.struct.types
                    prefix = name + ':'
                    get_value = lambda v, i: v[0][i]
                else:
                    group = {name: type}
                    prefix = ''
                    get_value = lambda v, i: v[i]
                for i, v in enumerate(group.items()):
                    n, t = v
                    joined_name = prefix + n
                    if 'complex' in t.kind:
                        ordinate = np.array([complex(*get_value(v, i)) for v in vals])
                    else:
                        ordinate = np.array([get_value(v, i) for v in vals])
                    signal = Signal(
                        name = joined_name,
                        ordinate = ordinate,
                        type = t,
                        access = t.name,
                        units = t.units,
                        meta = meta,
                    )
                    signals[joined_name] = signal
                del values[name]
        else:
            # no traces, this should be a DC op-point analysis dataset
            for name, value in values.items():
                assert len(value.values) == 1
                type = types[value.type]
                if type.struct:
                    for t, v in zip(type.struct.types.values(), value.values[0][0]):
                        n = f'{name}.{t.name}'
                        v = Quantity(v, unicode_units(t.units))
                        signal = Signal(
                            name = n,
                            ordinate = v,
                            type = t,
                            units = t.units,
                            meta = meta,
                        )
                        signals[n] = signal
                else:
                    v = Quantity(value.values[0][0], unicode_units(type.units))
                    signal = Signal(
                        name = name,
                        ordinate = v,
                        type = type,
                        access = type.name,
                        units = type.units,
                        meta = meta,
                    )
                    signals[name] = signal
        self.signals = signals

        if update_cache:
            self._write_cache(cache_filepath)

    def get_sweep(self, index=0):
        """
        Get Sweep

        index (int):
            PSF allows multiple sweeps (abscissas). You can use this argument to
            select the one you want. The default is 0.
        """
        if self.sweeps:
            return self.sweeps[index]

    def get_signal(self, name):
        """
        Get Signal

        name (string):
            Name of signal return.

        Raises UnknownSignal (subclass of Error) if the name given does not
        correspond to a known signal.
        """
        try:
            return self.signals[name]
        except KeyError:
            raise UnknownSignal(name, choices=self.signals.keys())

    def all_signals(self):
        """
        All Signals

        Iterates through all signals.
        """
        for signal in self.signals.values():
            yield signal

    def log_x(self, sweep=None):
        """
        Log X

        True if PSF is recommending a logarithmic x-axis.
        """
        if not sweep:
            sweep = self.get_sweep()
        return sweep.grid == 3

    def log_y(self, sweep=None):
        """
        Log Y

        True if PSF is recommending a logarithmic y-axis.
        """
        if not sweep:
            sweep = self.get_sweep()
        return sweep.grid == 3

    @staticmethod
    def units_to_unicode(units):
        """
        Transform ASCII units to Unicode

        units (str):
            The units to transform.
        """
        return unicode_units(units)

    @staticmethod
    def units_to_latex(units):
        """
        Transform ASCII units to Latex

        Note: this function is not yet implemented.

        units (str):
            The units to transform.
        """
        return units

    def _read_cache(self, cache_filepath):
        with open(cache_filepath, 'rb') as f:
            self.__dict__ = pickle.load(f)

    def _write_cache(self, cache_filepath):
        with open(cache_filepath, 'wb') as f:
            pickle.dump(self.__dict__, f, pickle.HIGHEST_PROTOCOL)



from setuptools import setup
from codecs import open

with open('README.rst', encoding='utf-8') as f:
    readme = f.read()

setup(
    name = 'psf_utils',
    version = '1.5.0',
    description = 'Cadence PSF file utilities',
    long_description = readme,
    long_description_content_type = 'text/x-rst',
    author = "Ken Kundert",
    author_email = 'psf_utils@nurdletech.com',
    url = 'https://psf_utils.readthedocs.io',
    download_url = 'https://github.com/kenkundert/psf_utils/tarball/master',
    license = 'GPLv3+',
    packages = 'psf_utils'.split(),
    entry_points = {'console_scripts': [
        'list-psf = psf_utils.list:list_signals',
        'show-psf = psf_utils.show:show_signals',
    ]},
    install_requires = """
        docopt
        inform>=1.19
        matplotlib
        numpy
        ply==3.4
        quantiphy
    """.split(),
    python_requires = '>=3.6',
    zip_safe = True,
    keywords = 'cadence spectre PSF simulation'.split(),
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Utilities',
        'Topic :: Scientific/Engineering',
    ],
)
